
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '38.0.0rc3',
    'branch_nick': 'master',
    'revision_id': 'c4df1306d192b518b38e023e783ea998c6c0dfb5',
    'revno': 13917
}
